library(plyr)
library(reshape2)
library(foreach)
   library(doMC); registerDoMC()
library(rjags)
    load.module("glm")
library(ggplot2)

# ------------------------------------------------------------
# Get data
# ------------------------------------------------------------

subjects = read.table("../data/data.csv",
    header = TRUE, sep = ",", na.strings = ".")
row.names(subjects) = subjects$Subject

# ------------------------------------------------------------
# Utility functions
# ------------------------------------------------------------

qw = function(...)
# qw(a, b, c) → c("a", "b", "c")
   {m = match.call(expand.dots = FALSE)
    char(m[["..."]])}

char = as.character
ss = subset

ordf = function(`_data`, ...)
# Similar to plyr::arrange, but preserves row names.
# Data = ordf(Data, FactorB, -FactorA, …)
   {m = match.call()
#    vl = eval(call("list", as.list(m)[c(-1, -2)]), `_data`, parent.frame())
#    `_data`[do.call(order, vl),]}
#        do.call(order, as.list(m)[c(-1, -2)],
#           envir = parent.frame())),]}
   `_data`[eval(as.call(c(quote(order), as.list(m)[c(-1, -2)])),
        `_data`, parent.frame()),]}

ifna = function(a, b)
    ifelse(is.na(a), b, a)

modnames = function(x, ...)
# modnames(mtcars, c = "carb", f = "disp") renames two columns
# without touching the others.
   `names<-`(x, ifna(
       `names<-`(names(c(...)), c(...))[names(x)],
       names(x)))

coda.vars = function(samp, str)
# coda.vars(samp, "rho") gets all the parameter names in the
# mcmc.list `samp` that begin with "rho[", as well as any single
# parameter named "rho".
    varnames(samp)[
        varnames(samp) == str |
        c(substr(varnames(samp), 1, nchar(str) + 1) == paste(sep = '', str, "["))]

as.bugs.model = function(expr)
# So you can put BUGS models in your source code without quoting
# them.
  {m = match.call()
   paste(sep = '', c("model ", paste(collapse = "\n", deparse(m$expr))))}

std.jags.rngs = c("base::Wichmann-Hill", "base::Marsaglia-Multicarry", "base::Super-Duper", "base::Mersenne-Twister")
multijags = function(
        modelfile,
        data, inits, monitor,
        n.adapt = 1000, n.nonadapt = 0,
        n.sample = 1000 * thin, thin = 1,
        quiet = F)
# We show progress bars only for chain 1.
   {if (!is.list(inits) || !is.list(inits[[1]]))
        stop("inits is not a list of lists")
    n.chains = length(inits)
    if (n.chains > length(std.jags.rngs))
        stop("Not enough RNGs for ", n.chains, " chains")
    results = foreach::foreach(i = 1 : n.chains) %dopar%
       {inits = inits[[i]]
        inits$.RNG.name = std.jags.rngs[[i]]
        jags = rjags::jags.model(modelfile, data, inits, n.chains = 1,
            n.adapt = 0, quiet = i != 1 | quiet)
        adapt(jags, n.iter = n.adapt, end.adaptation = T,
            progress.bar = ifelse(i == 1, "text", "none"))
        if (n.nonadapt)
           update(jags, n.iter = n.nonadapt,
               progress.bar = ifelse(i == 1, "text", "none"))
        samp = coda.samples(jags, variable.names = monitor,
            n.iter = n.sample, thin = thin,
            progress.bar = ifelse(i == 1, "text", "none"))
        list(jags, samp[[1]])}
    list(
        jags = lapply(results, function (l) l[[1]]),
        samp = do.call(mcmc.list, lapply(results, function (l) l[[2]])))}

mapcols = function(x, f) apply(x, 2, f)

no.gridlines = function() modifyList(theme_get(), list(
    panel.grid.major = theme_blank(),
    panel.grid.minor = theme_blank()))

interact = function(...)
    interaction(lex.order = T, ...)

groups = function(v)
# Given, say, a data frame D, groups(D$Foo) shows the values of
# factor Foo and how many cases have that value.
   {g = c(by(v, v, length))
    ifelse(is.na(g), 0, g)}

# ------------------------------------------------------------
# Processing
# ------------------------------------------------------------

sb = transform(subjects,
    target = factor(Target, labels = qw(direct, meta)),
    strategy = factor(Strategy, labels = qw(intuit, delib)))
sb$Subject = NULL

sb = modnames(sb,
    ses = "Session", strat = "Strategy")

qs = melt(subjects, id.var = "Subject",
    variable.name = "question", value.name = "est",
    measure.vars = qw(QA, QB, QC, QD, QE))
qs = modnames(qs, sn = "Subject")
#qs = cbind(qs, zlog =
#    dby(qs, .(question, target = sb[char(sn), "target"]),
#        z = zscore(log(est, 10)))$z)
#qs = transform(qs, sn = factor(sn, levels = char(unique(
#    ordf(dby(qs, .(sn), z = mean(zlog, na.rm = T)), z)$sn))))

# ------------------------------------------------------------
# MCMC
# ------------------------------------------------------------

model = as.bugs.model(
   {for (s in 1 : N.subjects)
       {for (q in 1 : N.questions)
           {y[q, s] <- exp(v[q, s])
            vhat[q, s] ~ dnorm(bq[q] + b.delib[q] * deliberate[s] + bs[s],
                v.s^(-2))
            v[q, s] ~ dnorm(bq[q] + b.delib[q] * deliberate[s] + bs[s],
                v.s^(-2))}
              # No constant term because it would create linear
              # dependence with the `bq`s. (Such a concern doesn't
              # apply to the `bs`es so much because they're random rather
              # than fixed effects.)
        bs[s] ~ dnorm(0, bs.s^(-2))}
    for (q in 1 : N.questions)
       {bq[q] ~ dnorm(0, 100^(-2))
        b.delib[q] ~ dnorm(0, 100^(-2))}
    v.s ~ dunif(0, 20)
    bs.s ~ dunif(0, 20)})

fake = within(list(),
   {v.s = .1
    bs.s = .1
    N.questions = 5
    bq = c(10, 10, 20, 30, 3)
    b.delib = c(-2, 1.5, -1, -.5, .25)
    N.subjects = 30
    deliberate = c(rep(0, N.subjects/2), rep(1, N.subjects/2))})

fake2 = within(list(),
   {v.s = .2
    bs.s = .001
    N.questions = 5
    bq = c(3, 10, 30, 10, 20)
    b.delib = c(-.5, .25, 1.5, -1, -2)
    N.subjects = 30
    deliberate = c(rep(0, N.subjects/2), rep(1, N.subjects/2))})

fake3 = within(list(),
   {v.s = .05
    bs.s = .5
    N.questions = 5
    bq = c(20, 10, 30, 3, 10)
    b.delib = c(-2, -1, .25, 1.5, -.5)
    N.subjects = 30
    deliberate = c(rep(0, N.subjects/2), rep(1, N.subjects/2))})

each.q = function (dat, f)
    outer(1 : dat$N.questions, 1 : dat$N.subjects, f)

set.seed(311)
gendata = function(dat)
   within(dat,
       {bs = rnorm(N.subjects, 0, bs.s)
        v = each.q(dat, function (q, s)
            rnorm(N.questions * N.subjects,
                bq[q] + b.delib[q] * deliberate[s] + bs[s],
                v.s))
        y = exp(v)})

fake = gendata(fake)

go.fake = function()
    with(fake, multijags(textConnection(model),
        data = list(
            N.questions = N.questions, N.subjects = N.subjects,
            deliberate = deliberate,
            v = log(y)),
        inits = lapply(list(fake, fake2, fake3),
                function (f) with(gendata(f),
                    list(v.s = v.s, bs.s = bs.s, bq = bq, b.delib = b.delib, bs = bs))),
        monitor = qw(vhat, v.s, bs.s, bq, b.delib, bs),
        n.adapt = 1000, n.nonadapt = 500, n.sample = 500))

reald = within(list(),
   {sb = ss(sb, target == "direct")
    qs = ss(qs, sn %in% row.names(sb))
    N.subjects = nrow(sb)
    N.questions = length(levels(qs$question))
    deliberate = sb$strategy == "delib"
    y = acast(qs, question ~ sn, value.var = "est")})

go = function()
    with(reald, multijags(textConnection(model),
        data = list(
            N.questions = N.questions, N.subjects = N.subjects,
            deliberate = deliberate,
            v = log(y)),
        inits = list(list(), list()),
        monitor = qw(vhat, v.s, bs.s, bq, b.delib, bs),
        n.adapt = 1000, n.nonadapt = 500, n.sample = 500))

ppc = function(dat, samp)
   {d = as.data.frame(t(mapcols(
        as.matrix(samp[[1]][, coda.vars(samp, "vhat") ]),
        function (v) quantile(v, c(.025, .25, .5, .75, .975)))))
    names(d) = qw(lo, midlo, mid, midhi, hi)
    d$strategy = ifelse(c(each.q(dat, function (q, s) dat$deliberate[s])),
        "delib", "intuit")
    d$q = factor(c(each.q(dat, function (q, s) q)))
    d$actual = log(c(dat$y))
    #d = ss(d, q == 5)
    d = ss(d, !is.na(actual))
    with(d, message(sprintf("The 95%% posterior intervals have %d%% coverage (%d of %d)",
        round(100 * mean(actual >= lo & actual <= hi)),
        sum(actual >= lo & actual <= hi),
        nrow(d))))
    with(d, message(sprintf("The 50%% posterior intervals have %d%% coverage (%d of %d)",
        round(100 * mean(actual >= midlo & actual <= midhi)),
        sum(actual >= midlo & actual <= midhi),
        nrow(d))))
    d = ordf(d, interact(strategy, q), actual)
    d$n = sequence(groups(interact(d$strategy, d$q)))
    ggplot(data = d) +
        geom_line(aes(n, lo), color = "blue") +
        geom_line(aes(n, hi), color = "blue") +
        geom_point(aes(n, actual)) +
        facet_grid(q ~ strategy, scales = "free") +
        ylab("predicted") + xlab("observation") +
        no.gridlines()}

# Try:
#   l = go()
#   ppc(reald, l$samp)
